/* function.cpp -- Copyright (c) 2004 Michael Pyne <michael.pyne@kdemail.net>
 * This software is licensed under the terms of the GNU General Public License,
 * v2 (or any later version).  This software is distributed WITH ABSOLUTELY NO
 * WARRANTY!.
 */
#include <kdebug.h>

#include <qvaluevector.h>
#include <qstring.h>
#include <qregexp.h>

#include <math.h>
#include "function.h"
#include "node.h"
#include "valuemanager.h"

// Used to try and avoid recursive function definitions
class DupFinder : public NodeFunctor
{
    public:
    DupFinder(const QString &nameToFind) :
        m_name(nameToFind), m_valid(true)
    {
    }

    virtual ~DupFinder() { }

    bool isValid() const { return m_valid; }

    virtual void operator()(const Node *node)
    {
	if(!m_valid)
	    return;

	const BaseFunction *fn = dynamic_cast<const BaseFunction *>(node);
	if(fn && fn->name() == m_name)
	    m_valid = false; // Duplicate detected
    }

    private:
    QString m_name;
    bool m_valid;
};

// The parser sets this variable at the end of a successful parse.
Abakus::number_t parse_result = 0.0;

// Used by the lastParseStatusResult() function
int _lastParseStatusResult = 0;

static Abakus::TrigMode m_trigMode = Abakus::Degrees;

// Define static member for FunctionManager
FunctionManager *FunctionManager::m_manager = 0;

FunctionManager *FunctionManager::instance()
{
    if(!m_manager)
	m_manager = new FunctionManager;

    return m_manager;
}

FunctionManager::FunctionManager(QObject *parent, const char *name) :
    QObject(parent, name)
{
    m_dict.setAutoDelete(true);
}

// Dummy return value to enable static initialization in the DECL_*()
// macros.
bool FunctionManager::addFunction(const QString &name, FunctionManager::function_t fn)
{
    Function *newFn = new Function;
    QRegExp returnTrigRE("^a(cos|sin|tan)");
    QRegExp needsTrigRE("^(cos|sin|tan)");
    QString fnName(name);

    newFn->name = name;
    newFn->fn = fn;
    newFn->userDefined = false;
    newFn->returnsTrig = fnName.contains(returnTrigRE);
    newFn->needsTrig = fnName.contains(needsTrigRE);

    m_dict.insert(name, newFn);

    return false;
}

// Declares a function, with the following parameters:
// * name - The user accessible name of the function
// * fn - The base name of the implementing function.
// * suffix - Suffix to add to the end of the function name
//   Example: f for float functions, l for long double functions, nothing for
//            double functions.
//
// Not only will the function be added to the function hash table, but an
// appropriate function prototype will be generated for broken system that
// don't do it for us.
#define DECLARE_FUNC(name, fn, suffix) Abakus::number_t fn##suffix(const Abakus::number_t); \
bool dummy##name = FunctionManager::instance()->addFunction(#name, fn##suffix)

// Declares a function defined by name, with fn == base name of the
// implementing function.
#define DECLARE_FUNC2(name, fn) DECLARE_FUNC(name, fn, l)

// Declares a function name that is implemented by the function of the
// same base name.
#define DECLARE_FUNC1(name) DECLARE_FUNC2(name, name)

DECLARE_FUNC1(sin);
DECLARE_FUNC1(cos);
DECLARE_FUNC2(abs, fabs);  // math.h version is called fabs
DECLARE_FUNC1(exp);
DECLARE_FUNC1(acos);
DECLARE_FUNC1(asin);
DECLARE_FUNC1(atan);
DECLARE_FUNC1(tan);
DECLARE_FUNC1(cosh);
DECLARE_FUNC1(sinh);
DECLARE_FUNC1(tanh);
DECLARE_FUNC2(ln, log);    // math.h version is called log
DECLARE_FUNC2(log, log10); // math.h version is called log10
DECLARE_FUNC1(sqrt);
DECLARE_FUNC1(ceil);
DECLARE_FUNC1(floor);

Function *FunctionManager::function(const QString &name)
{
    return m_dict[name];
}

// Returns true if the named identifier is a function, false otherwise.
bool FunctionManager::isFunction(const QString &name)
{
    return function(name) != 0;
}

bool FunctionManager::isFunctionUserDefined(const QString &name)
{
    const Function *fn = function(name);
    return (fn != 0) && (fn->userDefined);
}

bool FunctionManager::addFunction(BaseFunction *fn, const QString &dependantVar)
{
    // First see if this function is recursive
    DupFinder dupFinder(fn->name());
    UnaryFunction *unFunction = dynamic_cast<UnaryFunction *>(fn);
    if(unFunction && unFunction->operand()) {
	unFunction->operand()->applyMap(dupFinder);
	if(!dupFinder.isValid())
	    return false;
    }

    // Structure holds extra data needed to call the user defined
    // function.
    UserFunction *newFn = new UserFunction;
    newFn->sequenceNumber = m_dict.count();
    newFn->fn = fn;
    newFn->varName = QString(dependantVar);

    // Now setup the Function data structure that holds the information
    // we need to access and call the function later.
    Function *fnTabEntry = new Function;
    fnTabEntry->name = fn->name();
    fnTabEntry->userFn = newFn;
    fnTabEntry->returnsTrig = false;
    fnTabEntry->needsTrig = false;
    fnTabEntry->userDefined = true;

    if(m_dict.find(fn->name()))
	emit signalFunctionRemoved(fn->name());

    m_dict.replace(fn->name(), fnTabEntry);
    emit signalFunctionAdded(fn->name());

    return true;
}

void FunctionManager::removeFunction(const QString &name)
{
    Function *fn = function(name);

    // If we remove a function, we need to decrement the sequenceNumber of
    // functions after this one.
    if(fn && fn->userDefined) {
	int savedSeqNum = fn->userFn->sequenceNumber;

	// Emit before we actually remove it so that the info on the function
	// can still be looked up.
	emit signalFunctionRemoved(name);

	delete fn->userFn;
	fn->userFn = 0;
	m_dict.remove(name);

	QDictIterator<Function> it(m_dict);
	for (; it.current(); ++it) {
	    UserFunction *userFn = it.current()->userDefined ? it.current()->userFn : 0;
	    if(userFn && userFn->sequenceNumber > savedSeqNum)
		--it.current()->userFn->sequenceNumber;
	}
    }
}

QStringList FunctionManager::functionList(FunctionManager::FunctionType type)
{
    (void) type;
    QDictIterator<Function> it(m_dict);
    QStringList functions;

    switch(type) {
	case Builtin:
	    for(; it.current(); ++it)
		if(!it.current()->userDefined)
		    functions += it.current()->name;
	break;

	case UserDefined:
	    // We want to return the function names in the order they were
	    // added.
	    {
		QValueVector<Function *> fnTable(m_dict.count(), 0);
		QValueVector<int> sequenceNumberTable(m_dict.count(), -1);

		// First find out what sequence numbers we have.
		for(; it.current(); ++it)
		    if(it.current()->userDefined) {
			int id = it.current()->userFn->sequenceNumber;
			fnTable[id] = it.current();
			sequenceNumberTable.append(id);
		    }

		// Now sort the sequence numbers and return the ordered list
		qHeapSort(sequenceNumberTable.begin(), sequenceNumberTable.end());

		for(unsigned i = 0; i < sequenceNumberTable.count(); ++i)
		    if(sequenceNumberTable[i] >= 0)
			functions += fnTable[sequenceNumberTable[i]]->name;
	    }
	break;

	case All:
	    functions += functionList(Builtin);
	    functions += functionList(UserDefined);
	break;
    }

    return functions;
}

// Applies the function identified by func, using value as a parameter.
Abakus::number_t evaluateFunction(const Function *func, const Abakus::number_t value)
{
    Abakus::number_t tempValue(value), tempResult;
    Abakus::number_t semiCircle = trigMode() == Abakus::Degrees ? 180.0 : M_PI;

    if(func->userDefined) {
	// Pull real entry from userFunctionTable
	UserFunction *realFunction = func->userFn;
	ValueManager::instance()->setValue(realFunction->varName, tempValue);
	return realFunction->fn->value();
    }

    if(func->needsTrig && trigMode() == Abakus::Degrees)
	tempValue *= (M_PI / 180.0);

    // Special case the tan function
    if(func->fn == tanl && fabsl(fmodl(value, semiCircle)) == (semiCircle / 2))
	return NAN;

    // Special case sin
    if(func->fn == sinl && fabsl(fmodl(value, semiCircle)) == 0.0)
	return 0.0;

    // Special case cos
    if(func->fn == cosl && fabsl(fmodl(value - semiCircle / 2, semiCircle)) == 0.0)
	return 0.0;

    tempResult = func->fn(tempValue);

    if(func->returnsTrig && trigMode() == Abakus::Degrees)
	tempResult *= (180.0 / M_PI);

    return tempResult;
}

// Returns the status of the last parse operator.  non-zero indicates error,
// 0 indicates success.
int lastParseStatusResult(void)
{
    return _lastParseStatusResult;
}

void setTrigMode(Abakus::TrigMode mode)
{
    m_trigMode = mode;
}

Abakus::TrigMode trigMode()
{
    return m_trigMode;
}
