#ifndef _ABAKUS_NODE_H
#define _ABAKUS_NODE_H

#include <qstring.h>

#include "numerictypes.h"

class Node;
class Function;

/**
 * A class that operates on a Node.  Called recursively on a node and all
 * of its children.
 */
class NodeFunctor
{
    public:
    virtual void operator()(const Node *node) = 0;
    virtual ~NodeFunctor() { }
};

class Node
{
    public:
    virtual ~Node() { }
    virtual Abakus::number_t value() const = 0;

    // Deletes a node only if it isn't a function, since those are
    // typically read-only.
    virtual void deleteNode(Node *node);

    // Creates a perfect replica of the object, must be implemented in
    // all subclasses.
    virtual Node *clone() const = 0;

    // Calls functor() on all subchildren and this.
    virtual void applyMap(NodeFunctor &fn) const = 0;

    // Returns an infix representation of the expression.
    virtual QString infixString() const = 0;
};

class BaseFunction : public Node
{
    public:
    BaseFunction(const char *name);

    virtual QString name() const { return m_name; }
    virtual const Function *function() const { return m_function; }

    private:
    QString m_name;
    const Function *m_function;
};

class UnaryFunction : public BaseFunction
{
    public:
    UnaryFunction(const char *name, Node *operand);
    virtual ~UnaryFunction();

    virtual Node *operand() const { return m_node; }
    virtual void setOperand(Node *operand);

    virtual void applyMap(NodeFunctor &fn) const;
    virtual QString infixString() const;

    private:
    Node *m_node;
};

class UserDefinedFunction : public UnaryFunction
{
    public:
    UserDefinedFunction(const char *name, Node *operand) : UnaryFunction(name, operand) { };

    virtual Abakus::number_t value() const { return operand()->value(); }

    virtual UserDefinedFunction *clone() const;
};

class BuiltinFunction : public UnaryFunction
{
    public:
    BuiltinFunction(const char *name, Node *operand);

    virtual BuiltinFunction *clone() const;

    virtual Abakus::number_t value() const;
};

class UnaryOperator : public Node
{
    public:
    typedef enum { Negation } Type;

    UnaryOperator(Type type, Node *operand);
    virtual ~UnaryOperator();

    virtual UnaryOperator *clone() const;
    virtual void applyMap(NodeFunctor &fn) const;
    virtual QString infixString() const;

    virtual Abakus::number_t value() const;

    virtual Type type() const { return m_type; }
    virtual Node *operand() const { return m_node; }

    private:
    Type m_type;
    Node *m_node;
};

class BinaryOperator : public Node
{
    public:
    typedef enum { Addition, Subtraction, Multiplication, Division, Exponentiation } Type;

    BinaryOperator(Type type, Node *left, Node *right);
    virtual ~BinaryOperator();

    virtual BinaryOperator *clone() const;
    virtual void applyMap(NodeFunctor &fn) const;
    virtual QString infixString() const;

    virtual Abakus::number_t value() const;

    virtual Type type() const       { return m_type; }
    virtual Node *leftNode() const  { return m_left; }
    virtual Node *rightNode() const { return m_right; }

    private:
    bool isSimpleNode(Node *node) const;

    Type m_type;
    Node *m_left, *m_right;
};

class Identifier : public Node
{
    public:
    Identifier(const char *name);

    virtual Identifier *clone() const { return new Identifier(name()); }
    virtual void applyMap(NodeFunctor &fn) const;
    virtual QString infixString() const { return name(); }

    virtual Abakus::number_t value() const;

    virtual QString name() const { return m_name; }

    private:
    QString m_name;
};

class NumericValue : public Node
{
    public:
    NumericValue(const Abakus::number_t value) : m_value(value) { }

    virtual NumericValue *clone() const { return new NumericValue(value()); }
    virtual void applyMap(NodeFunctor &fn) const { fn(this); }
    virtual QString infixString() const { return QString::number(value(), 'g', 18); }

    virtual Abakus::number_t value() const { return m_value; }

    private:
    Abakus::number_t m_value;
};

#endif

// vim: set et sw=4 ts=8:
