/* parser.y -- Bison definition file for the parser for qt-calc.
 * By Michael Pyne <michael.pyne@kdemail.net>
 * Copyright (c) 2004.
 *
 * You may distribute and modify this file under the terms of the GNU
 * General Public License, v2 (or later).  Notably, this program is distributed
 * WITH ABSOLUTELY NO WARRANTY.
 */

%{

/* Add necessary includes here. */
#include <kdebug.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "node.h"
#include "function.h"
#include "error_handler.h"
#include "valuemanager.h"

extern char *yytext;

int __parserResult;

int yylex(void);
int yyerror(const char *);

%}

%union {
    Node *node;
    NumericValue *value;
    UnaryFunction *fn;
    Identifier *ident;
}

%token <value> NUM
%type <node> EXP FACTOR TERM S EXPONENT NUMBER FINAL
%type <fn> FUNC
%token <fn> FN
%token <ident> ID
%type <ident> IDENT ASSIGN
%token POWER "**"
%token SET "set"
%token REMOVE "remove"

%%

FINAL: S {
    parse_result = $1->value();
    delete $1;
    $$ = 0;
}

S: EXP { $$ = $1; }

S: error '=' {
    post_error_message("This is an invalid assignment%1.", "");
    YYABORT;
}

S: FUNC '=' {
    post_error_message("You can't assign to function %1.", $1->name());
    YYABORT;
}

ASSIGN: '(' IDENT ')' '=' {
    if(!ValueManager::instance()->isValueSet($2->name()))
	ValueManager::instance()->setValue($2->name(), 0.0);

    $$ = $2;
}

S: SET FUNC ASSIGN EXP {
    // We're trying to reassign an already defined function, make sure it's
    // not a built-in.
    QString funcName = $2->name();
    QString ident = $3->name();
    FunctionManager *manager = FunctionManager::instance();

    if(manager->isFunctionUserDefined(funcName)) {
	post_error_message("Function %1 is built-in and cannot be overridden.", funcName);
	YYABORT;
    }

    BaseFunction *newFn = new UserDefinedFunction(funcName, $4);
    if(!manager->addFunction(newFn, ident)) {
	post_error_message("Unable to define function %1 because it is recursive", funcName);
	YYABORT;
    }

    __parserResult = 2;
    $$ = new NumericValue(1.0);
}

S: SET IDENT ASSIGN EXP {
    QString funcName = $2->name();
    QString ident = $3->name();

    // No need to check if the function is already defined, because the
    // lexer checked for us before returning the IDENT token.
    BaseFunction *newFn = new UserDefinedFunction(funcName, $4);
    FunctionManager::instance()->addFunction(newFn, ident);

    __parserResult = 2;
    $$ = new NumericValue(1.0);
}

S: SET IDENT '=' EXP {
    ValueManager::instance()->setValue($2->name(), $4->value());

    $$ = $4;
    __parserResult = 2;
}

S: REMOVE FUNC '(' ')' {
    FunctionManager::instance()->removeFunction($2->name());
    $$ = new NumericValue(1.0);
    __parserResult = 2;
}

S: REMOVE IDENT '(' ')' {
    // This is an error
    post_error_message("Function %1 is not defined.", $2->name());
    YYABORT;
}

S: REMOVE IDENT '(' IDENT ')' {
    // This is an error
    post_error_message("Function %1 is not defined.", $2->name());
    YYABORT;
}

S: REMOVE IDENT {
    ValueManager *manager = ValueManager::instance();
    
    if(manager->isValueSet($2->name()) && !manager->isValueReadOnly($2->name())) {
	manager->removeValue($2->name());
	$$ = new NumericValue(1.0);
	__parserResult = 2;
    }
    else {
	if(manager->isValueSet($2->name()))
	    post_error_message("Can't remove predefined variable %1.", $2->name());
	else
	    post_error_message("Can't remove undefined variable %1.", $2->name());

	YYABORT;
    }
}

S: IDENT '=' EXP {
    ValueManager::instance()->setValue($1->name(), $3->value());

    $$ = $3;
    __parserResult = 2;
}

S: error {
    post_error_message("Sorry, I can't figure it out.%1", "");
    YYABORT;
}

EXP: EXP '+' FACTOR { $$ = new BinaryOperator(BinaryOperator::Addition, $1, $3); }
EXP: EXP '-' FACTOR { $$ = new BinaryOperator(BinaryOperator::Subtraction, $1, $3); }
EXP: FACTOR { $$ = $1; }

FACTOR: FACTOR '*' EXPONENT { $$ = new BinaryOperator(BinaryOperator::Multiplication, $1, $3); }
FACTOR: FACTOR '/' EXPONENT { $$ = new BinaryOperator(BinaryOperator::Division, $1, $3); }
FACTOR: EXPONENT { $$ = $1; }

EXPONENT: TERM POWER EXPONENT { $$ = new BinaryOperator(BinaryOperator::Exponentiation, $1, $3); }
EXPONENT: TERM { $$ = $1; }

TERM: '+' NUMBER { $$ = $2; }
TERM: '+' IDENT { $$ = $2; }
TERM: '-' NUMBER { $$ = new UnaryOperator(UnaryOperator::Negation, $2); }
TERM: '-' IDENT { $$ = new UnaryOperator(UnaryOperator::Negation, $2); }
TERM: '(' EXP ')' { $$ = $2; }
TERM: '-' '(' EXP ')' { $$ = new UnaryOperator(UnaryOperator::Negation, $3); }

TERM: NUMBER { $$ = $1; }

NUMBER: NUM {
    char *endptr = 0;
    Abakus::number_t value = strtod(yytext, &endptr);
    if(endptr == yytext) {
	post_error_message("Unable to convert %1 to a number!", QString(yytext));
	YYABORT;
    }

    $$ = new NumericValue(value);
}

TERM: FUNC TERM {
    $1->setOperand($2);
    $$ = $1;
}

TERM: IDENT {
    if(ValueManager::instance()->isValueSet($1->name()))
	$$ = $1;
    else {
	post_error_message ("Unknown variable %1", $1->name());
	YYABORT;
    }
}

IDENT: ID {
    $$ = new Identifier(yytext);
}

FUNC: FN {
    /* No check necessary, the lexer has already checked for us. */
    $$ = new BuiltinFunction(yytext, 0);
}
