#include <math.h>

#include <kdebug.h>
#include <klocale.h>

#include <qvaluestack.h>
#include <qregexp.h>
#include <qstringlist.h>

#include "rpnmuncher.h"
#include "valuemanager.h"
#include "function.h"

/**
 * Holds either a textual identifier, or a numeric value.
 */
class Operand
{
    public:
    Operand() : m_isValue(true), m_value(0.0) { }
    Operand(const QString &ident) : m_isValue(false), m_text(ident) { }
    Operand(Abakus::number_t value) : m_isValue(true), m_value(value) { }

    Abakus::number_t value() const
    {
	if(m_isValue)
	    return m_value;

	return ValueManager::instance()->value(m_text);
    }

    operator Abakus::number_t() const
    {
	return value();
    }

    QString text() const { return m_text; }

    private:
    bool m_isValue;
    QString m_text;
    Abakus::number_t m_value;
};

typedef QValueStack<Operand> OperandStack;

typedef enum { Number = 256, Func, Ident, Power, Set, Remove, Unknown } Token;

static int tokenize (const QString &token);

QString RPNParser::m_errorStr;
bool RPNParser::m_error(false);

Abakus::number_t RPNParser::rpnParseString(const QString &text)
{
    QStringList tokens = QStringList::split(QRegExp("\\s"), text);
    OperandStack stack;

    // Used in the case statements below
    Operand l, r;
    FunctionManager *manager = FunctionManager::instance();
    Function *fn = 0;

    m_error = false;
    m_errorStr = QString::null;

    for(QStringList::ConstIterator it = tokens.begin(); it != tokens.end(); ++it) {
	switch(tokenize(*it)) {
	    case Number:
	        stack.push((*it).toDouble());
	    break;

	    case Func:
		if(stack.count() < 1) {
		    m_error = true;
		    m_errorStr = i18n("Insufficient operands for function %1").arg(*it);
		    return 0.0;
		}

		fn = manager->function(*it);

		l = stack.pop();
		stack.push(evaluateFunction(fn, l));
	    break;

	    case Ident:
	        stack.push(*it);
	    break;

	    case Set:
	    case Remove:
	        m_error = true;
		m_errorStr = i18n("The set and remove commands can only be used in normal mode.");
		return 0.0;
	    break;

	    case Power:
	        if(stack.count() < 2) {
		    m_error = true;
		    m_errorStr = i18n("Insufficient operands for exponentiation operator.");
		    return 0.0;
		}

		r = stack.pop();
		l = stack.pop();
		stack.push(powl(l, r));
	    break;

	    case Unknown:
	        m_error = true;
		m_errorStr = i18n("Unknown token %1").arg(*it);
		return 0.0;
	    break;

	    case '=':
	        r = stack.pop();
		l = stack.pop();
		ValueManager::instance()->setValue(l.text(), r);

		stack.push(l);
	    break;

	    case '+':
	        if(stack.count() < 2) {
		    m_error = true;
		    m_errorStr = i18n("Insufficient operands for addition operator.");
		    return 0.0;
		}

		r = stack.pop();
		l = stack.pop();
		stack.push(l + r);
	    break;

	    case '-':
	        if(stack.count() < 2) {
		    m_error = true;
		    m_errorStr = i18n("Insufficient operands for subtraction operator.");
		    return 0.0;
		}

		r = stack.pop();
		l = stack.pop();
		stack.push(l - r);
	    break;

	    case '*':
	        if(stack.count() < 2) {
		    m_error = true;
		    m_errorStr = i18n("Insufficient operands for multiplication operator.");
		    return 0.0;
		}

		r = stack.pop();
		l = stack.pop();
		stack.push(l * r);
	    break;

	    case '/':
	        if(stack.count() < 2) {
		    m_error = true;
		    m_errorStr = i18n("Insufficient operands for division operator.");
		    return 0.0;
		}

		r = stack.pop();
		l = stack.pop();
		stack.push(l / r);
	    break;

	    default:
	        // Impossible case happened.
		kdError() << "Impossible case happened in " << k_funcinfo << endl;
		m_error = true;
		m_errorStr = "Bug found in program, please report.";
		return 0.0;
	}
    }

    // Make sure there is exactly one entry on stack
    if(stack.count() != 1) {
	m_error = true;
	m_errorStr = i18n("Invalid expression");
	return 0.0;
    }

    return stack.pop();
}

static int tokenize (const QString &token)
{
    bool isOK;

    token.toDouble(&isOK);
    if(isOK)
	return Number;

    if(token == "**" || token == "^")
	return Power;

    if(FunctionManager::instance()->isFunction(token))
	return Func;

    if(token.lower() == "set")
	return Set;

    if(token.lower() == "remove")
	return Remove;

    if(QRegExp("^\\w+$").search(token) != -1 &&
       QRegExp("\\d").search(token) == -1)
    {
	return Ident;
    }

    if(QRegExp("^[-+*/=]$").search(token) != -1)
	return token[0];

    return Unknown;
}
